/*
 * sturmbahnfahrer
 * (c) 2006 by Bram Stolk
 * bram at gmail.com
 * LICENSED ACCORDING TO THE GPL
 */

#include "dynamicobject.h"
#include "carobject.h"

#define LENGTH 1.0      // chassis length
#define WIDTH  0.6      // chassis width
#define HEIGHT 0.40     // chassis height
#define CMASS 4         // chassis mass
#define WMASS 0.20      // wheel mass
#define WHEELW 0.2      // wheel width
#define WHEELRAD 0.2    // wheel radius


#define SINK  0.17	// How much do we sink the center of mass?

CarObject::CarObject
(
  ssgEntity *bodymodel,
  ssgEntity *wheelmodel,
  ssgEntity *licplatemodel,
  dWorldID world, 
  dSpaceID bigspace,
  sgVec3 initialpos
) :
WorldObject(bodymodel),
brakelightstate(0),
accelerator(0),
brake(0),
steer(0),
downtime(0),
stationarytime(0),
action(false),
reversegear(false),
ebrake(false),
ground_object(0),
snapshot_available(false)
{
  name = "car";
  brakelightstate = FindState("brakelight", bodymodel);
  // load visible model as plib entity
  ssgBranch *branch = new ssgBranch();
  chassis_trf = new ssgTransform();
  chassis_trf->setName("chassis_trf");
  sink_trf = new ssgTransform();
  sink_trf->setName("sink_trf");
  sgVec3 sinkvec;
  sgSetVec3(sinkvec,0,0,SINK);
  sink_trf->setTransform(sinkvec);
  chassis_trf->addKid(sink_trf);
  branch->addKid(chassis_trf);
  sink_trf->addKid(bodymodel);
  if (licplatemodel) sink_trf->addKid(licplatemodel);
  int i;
  for (i=0; i<4; i++)
  {
    wheel_trfs[i] = new ssgTransform();
    branch->addKid(wheel_trfs[i]);
    wheel_trfs[i]->addKid(wheelmodel);
  }
  entity = branch;

  // chassis
  chassis_body = dBodyCreate(world);
  dBodySetAutoDisableFlag(chassis_body, false);
  dBodySetPosition (chassis_body,initialpos[0],initialpos[1],initialpos[2]);
  dMass m;
  dMassSetBox (&m,1,LENGTH,WIDTH,HEIGHT);
  dMassAdjust (&m,CMASS);
  dBodySetMass (chassis_body,&m);


  chassis_geom = dCreateBox (0,LENGTH,WIDTH,HEIGHT);
  dGeomSetData (chassis_geom,this);
#ifndef SINK
  dGeomSetBody (chassis_geom,chassis_body);
#endif


#ifdef SINK
  transform_geom = dCreateGeomTransform(0);
  dGeomSetData (transform_geom,this);
  dGeomSetBody (transform_geom,chassis_body);
  dGeomTransformSetInfo(transform_geom, 1);
  dGeomTransformSetGeom(transform_geom, chassis_geom);
  dGeomSetPosition(transform_geom, initialpos[0],initialpos[1],initialpos[2]);
  dGeomSetPosition(chassis_geom, 0,0,SINK);
  dBodySetPosition(chassis_body,initialpos[0],initialpos[1],initialpos[2]-SINK);
#endif

  // wheel bodies
  for (i=0; i<4; i++)
  {
    wheel_bodies[i] = dBodyCreate(world);
    dBodySetAutoDisableFlag(wheel_bodies[i], false);
    dQuaternion q;
    dQFromAxisAndAngle (q,1,0,0,M_PI*0.5);
    dBodySetQuaternion (wheel_bodies[i],q);
    dMassSetSphere (&m,1,WHEELRAD);
    dMassAdjust (&m,WMASS);
    dBodySetMass (wheel_bodies[i],&m);
    if (i<=2) dBodySetFiniteRotationMode(wheel_bodies[i], true); // Added, BSt
    wheel_geoms[i] = dCreateCylinder(0, WHEELRAD, WHEELW);
    dGeomSetBody (wheel_geoms[i],wheel_bodies[i]);
    dGeomSetData (wheel_geoms[i],this);
  }
  dBodySetPosition (wheel_bodies[0],initialpos[0]+0.45*LENGTH,initialpos[1]+WIDTH*0.60,initialpos[2]-HEIGHT*0.50);
  dBodySetPosition (wheel_bodies[1],initialpos[0]+0.45*LENGTH,initialpos[1]-WIDTH*0.60,initialpos[2]-HEIGHT*0.50);
  dBodySetPosition (wheel_bodies[2],initialpos[0]-0.45*LENGTH,initialpos[1]+WIDTH*0.60,initialpos[2]-HEIGHT*0.50);
  dBodySetPosition (wheel_bodies[3],initialpos[0]-0.45*LENGTH,initialpos[1]-WIDTH*0.60,initialpos[2]-HEIGHT*0.50);

  // front and back wheel hinges
  for (i=0; i<4; i++)
  {
    joint[i] = dJointCreateHinge2 (world,0);
    dJointAttach (joint[i],chassis_body,wheel_bodies[i]);
    const dReal *a = dBodyGetPosition (wheel_bodies[i]);
    dJointSetHinge2Anchor (joint[i],a[0],a[1],a[2]);
    dJointSetHinge2Axis1 (joint[i],0,0,1);
    dJointSetHinge2Axis2 (joint[i],0,1,0);
  }

  // set joint suspension
  for (i=0; i<4; i++)
  {
//    dJointSetHinge2Param (joint[i],dParamSuspensionERP,0.35);
//    dJointSetHinge2Param (joint[i],dParamSuspensionCFM,0.02);
    dJointSetHinge2Param (joint[i],dParamSuspensionERP,0.25);
    dJointSetHinge2Param (joint[i],dParamSuspensionCFM,0.05);
  }

  // lock back wheels along the steering axis
  for (i=2; i<=3; i++)
  {
    // set stops to make sure wheels always stay in alignment
    dJointSetHinge2Param (joint[i],dParamLoStop,0);
    dJointSetHinge2Param (joint[i],dParamHiStop,0);
    // Do this twice... weird ode bug! see mailing list and Jon's carworld
    dJointSetHinge2Param (joint[i],dParamLoStop,0);
    dJointSetHinge2Param (joint[i],dParamHiStop,0);
    dJointSetHinge2Param (joint[i],dParamStopERP,0.99);
    dJointSetHinge2Param (joint[i],dParamStopCFM,0.01);
  }

  space = dSimpleSpaceCreate(bigspace);
  dSpaceSetCleanup(space,0);
#ifndef SINK
  dSpaceAdd (space,chassis_geom);
#else
  dSpaceAdd (space,transform_geom);
#endif
  for (i=0; i<4; i++)
    dSpaceAdd (space,wheel_geoms[i]);
}


CarObject::~CarObject()
{
  int i;
  for (i=0; i<4; i++)
  {
    dGeomDestroy(wheel_geoms[i]);
    dJointDestroy(joint[i]);
  }
  dGeomDestroy(chassis_geom);
  dGeomDestroy(transform_geom);
  dSpaceDestroy(space);
}


float CarObject::GetWheelVelocity(void)
{
  float vel0 = fabs(*dBodyGetAngularVel(wheel_bodies[2]));
  float vel1 = fabs(*dBodyGetAngularVel(wheel_bodies[3]));
  return std::max(vel0, vel1);
}


bool CarObject::IsStationary(void) const
{
  // Are we stationary?
  const dReal *vel = dBodyGetLinearVel(chassis_body);
  dReal v[3]={vel[0],vel[1],vel[2]};

  if (ground_object)
  {
    const dReal *gvel = dBodyGetLinearVel(ground_object->body);
    v[0] -= gvel[0];
    v[1] -= gvel[1];
    v[2] -= gvel[2];
  }

  dReal magnitude = dSqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
  bool stationary = magnitude < 0.18;

  // HACK: Turn Table has rotational velocity, which is hard to interpret.
  // We simply assume we are stationary, so reverse gear is not blocked.
  if (ground_object)
  { 
    if (ground_object->name == "turntablewheel")
      stationary = true;
  }

  return stationary;
}


void CarObject::Simulate(float dt)
{
  // Get velocity of rear wheels.
  dReal v0 = -dJointGetHinge2Param (joint[2],dParamVel2);
  dReal v1 = -dJointGetHinge2Param (joint[2],dParamVel2);
  dReal currentv = (v0+v1)/2;

  dReal steerv = steer * 0.6;
  dReal speedv = 0.0;
  dReal fmax   = 0.0;

  // Are we in reverse gear?
  if (IsStationary())
    stationarytime += dt;
  else
    stationarytime = 0;
  if (stationarytime>0.15 && brake>0)
    reversegear=true;
  if (brake==0)
    reversegear=false;

  // Are we free rolling?
  if (brake==0 && accelerator==0 && ebrake==false)
  { 
    // apply a little engine braking
    speedv = 0.90 * currentv;
    fmax = 0.5;
  }
  else
  {
    // Are we braking?
    if ((brake>0 && !reversegear) || (accelerator>0 && reversegear) || ebrake)
    {
      speedv = 0;
      fmax = 50;
    }
    else
    {
      if (reversegear)
      {
        speedv = -70.0 * brake;
        fmax = 1.0 + 3.0 * brake;
      }
      else
      {
        speedv =  70.0 * accelerator;
        fmax = 1.0 + 3.0 * accelerator;
      }
    }
  }

  // motor: drive rear wheels.
  for (int jnr=2; jnr<=3; jnr++)
  {
    dJointSetHinge2Param (joint[jnr],dParamVel2,-speedv);
    dJointSetHinge2Param (joint[jnr],dParamFMax2,fmax);
  }

  // steering: turn front wheels.
  dReal curr0 = dJointGetHinge2Angle1(joint[0]);
  dReal curr1 = dJointGetHinge2Angle1(joint[1]);
  v0 = steerv - curr0;
  v1 = steerv - curr1;
  if (v0 > 0.1) v0 = 0.1;
  if (v0 < -0.1) v0 = -0.1;
  v0 *= 30.0;
  if (v1 > 0.1) v1 = 0.1;
  if (v1 < -0.1) v1 = -0.1;
  v1 *= 30.0;
  dJointSetHinge2Param (joint[0],dParamVel,v0);
  dJointSetHinge2Param (joint[0],dParamFMax,95);
  dJointSetHinge2Param (joint[0],dParamLoStop,-0.75);
  dJointSetHinge2Param (joint[0],dParamHiStop,0.75);
  dJointSetHinge2Param (joint[0],dParamFudgeFactor,0.1);
  dJointSetHinge2Param (joint[1],dParamVel,v1);
  dJointSetHinge2Param (joint[1],dParamFMax,95);
  dJointSetHinge2Param (joint[1],dParamLoStop,-0.75);
  dJointSetHinge2Param (joint[1],dParamHiStop,0.75);
  dJointSetHinge2Param (joint[1],dParamFudgeFactor,0.1);

  dBodySetFiniteRotationAxis(wheel_bodies[0],0,0,1);
  dBodySetFiniteRotationAxis(wheel_bodies[1],0,0,1);
}


void CarObject::Recover(float dt)
{
  sgVec3 carz;
  sgMat4 mat;
  chassis_trf->getTransform(mat);
  sgCopyVec3(carz, mat[2]);
  if (carz[2]<0.3)
  {
    downtime += dt;
  }
  else
  {
    downtime = 0.0;
  }
  if (downtime>2.8)
  {
    static float sign=1.0;
    float magnitude = 3.0 / dt;
    sgVec3 f;
    sgScaleVec3(f, carz, sign*magnitude);
    downtime=0.0;
    dBodyAddForce(chassis_body, 0,0,2*magnitude);
    dBodyAddForce(wheel_bodies[0],  f[0], f[1], f[2]);
    dBodyAddForce(wheel_bodies[2],  f[0], f[1], f[2]);
    dBodyAddForce(wheel_bodies[1], -f[0],-f[1],-f[2]);
    dBodyAddForce(wheel_bodies[3], -f[0],-f[1],-f[2]);
    sign=-sign;
  }
}


void CarObject::Sustain(float dt)
{
  Simulate(dt);
  Recover(dt);
  SetTransformFromBody(chassis_trf, chassis_body);
  for (int i=0; i<4; i++)
    SetTransformFromBody(wheel_trfs[i], wheel_bodies[i]);
  SetBrakeLight();
}


ssgSimpleState *CarObject::FindState(const std::string &statename, ssgEntity *ent)
{
  if (!ent) return 0;
  if (ent->isAKindOf(ssgTypeLeaf()))
  {
    ssgLeaf *leaf = (ssgLeaf*) ent;
    ssgState *st = leaf->getState();
    if (st->getName() == statename && st->isAKindOf(ssgTypeSimpleState()))
      return (ssgSimpleState*) st;
  }
  if (ent->isAKindOf(ssgTypeBranch()))
  {
    ssgBranch *branch = (ssgBranch*) ent;
    for (int i=0; i<branch->getNumKids(); i++)
    {
      ssgEntity *kid = branch->getKid(i);
      assert(kid);
      ssgSimpleState *state = FindState(statename, kid);
      if (state) return state;
    }
  }
  return 0;
}


void CarObject::SetBrakeLight(void)
{
  if (!brakelightstate) return;
  if (reversegear)
  {
    brakelightstate->setMaterial(GL_EMISSION, 1,1,1,1);
    return;
  }
  if (ebrake || brake>0.0f)
  {
    brakelightstate->setMaterial(GL_EMISSION, 1,0,0,1);
    return;
  }
  if (!reversegear && !ebrake && !brake)
  {
    brakelightstate->setMaterial(GL_EMISSION, 0,0,0,1);
  }
}


bool CarObject::SnapShot(bool revert)
{
  static dVector3 cpos;
  static dQuaternion crot;
  static dVector3 wpos[4];
  static dQuaternion wrot[4];

  if (revert)
  {
    if (!snapshot_available) return false;
    dBodySetPosition (chassis_body, cpos[0], cpos[1], cpos[2]);
    dBodySetQuaternion (chassis_body, crot);
    dBodySetLinearVel (chassis_body, 0,0,0);
    dBodySetAngularVel (chassis_body, 0,0,0);
    for (int i=0; i<4; i++)
    {
      dBodySetPosition (wheel_bodies[i], wpos[i][0], wpos[i][1], wpos[i][2]);
      dBodySetQuaternion (wheel_bodies[i], wrot[i]);
      dBodySetLinearVel (wheel_bodies[i], 0,0,0);
      dBodySetAngularVel (wheel_bodies[i], 0,0,0);
    }
  }
  else
  {
    memcpy(cpos, dBodyGetPosition(chassis_body), sizeof(cpos));
    memcpy(crot, dBodyGetQuaternion(chassis_body), sizeof(crot));
    for (int i=0; i<4; i++)
    {
      memcpy(wpos[i], dBodyGetPosition(wheel_bodies[i]), sizeof(wpos[i]));
      memcpy(wrot[i], dBodyGetQuaternion(wheel_bodies[i]), sizeof(wrot[i]));
    }
    snapshot_available = true;
  }
  return true;
}

